 /*
 * A driver for PCMCIA CardMan Mobile 4000
 *
 * cardman_cs.c v1.0 Tue Oct 10 11:56:46 MET 2000 support.linux@omnikey.com
 *                   
 *
 * Tue Oct 23 11:32:43 GMT 2001 herp - cleaned up header files
 * Sun Jan 20 10:11:15 MET 2002 herp - added modversion header files
 * Thu Nov 14 16:34:11 GMT 2002 mh   - added PPS functionality
 * Tue Nov 19 16:36:27 GMT 2002 mh   - added SUSPEND/RESUME functionailty
 *
 * current version: 2.3.1
 *
 * (C) 2000,2001,2002,2003 Omnikey AG
 *
 * All rights reserved
 */

#undef	PCMCIA_DEBUG
#undef	CHECK_VERSION

#include	<linux/config.h>
#include	<linux/version.h>

#ifdef CONFIG_MODVERSIONS
#define	MODVERSIONS
#include	<linux/modversions.h>
#endif

#include	<linux/kernel.h>
#include	<linux/module.h>
#include	<linux/slab.h>
#include        <linux/init.h>
#include	<linux/fs.h>
#include        <linux/delay.h>
#include	<asm/uaccess.h>
#include	<asm/io.h>

#include	<pcmcia/version.h>
#include	<pcmcia/cs_types.h>
#include	<pcmcia/cs.h>
#include	<pcmcia/cistpl.h>
#include	<pcmcia/cisreg.h>
#include	<pcmcia/ciscode.h>
#include	<pcmcia/ds.h>
#include	<pcmcia/bus_ops.h>

#define		__CM4000__
#include	"../cardman.h"

#undef	min
#define	min(a,b)	(a<b?a:b)

#define PCMCIA_DEBUG 0

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
static int io=0;
MODULE_PARM(io, "i");
#define DEBUG(n, args...) if (pc_debug >= (n)) printk(KERN_INFO args)
#else
#define DEBUG(n, args...)
#endif
static volatile char *version =
"cardman_cs.c v2.3.1 2003/04/24 support.linux@omnikey.com";

#define	T_1SEC		(HZ)
#define	T_10MSEC	(HZ/100)
#define	T_20MSEC	(HZ/50)
#define	T_25MSEC	(HZ/40)
#define	T_40MSEC	(HZ/25)
#define	T_50MSEC	(HZ/20)
#define	T_100MSEC	(HZ/10) 
#define	T_500MSEC	(HZ/2)

/*fops-protos:*/

static loff_t cmm_llseek(struct file *,loff_t,int);
static ssize_t cmm_read(struct file *,char *,size_t,loff_t *);
//static ssize_t cmm_write(struct file *,const char *,size_t,loff_t *);
static int cmm_write(struct file *,const char *,size_t,loff_t *);
static int cmm_ioctl(struct inode *,struct file *,unsigned int,unsigned long);
static int cmm_open(struct inode *,struct file *);
static int cmm_close(struct inode *,struct file *);

static struct file_operations cardman_fops = {
	llseek:		cmm_llseek,
	read:		cmm_read,
	write:		cmm_write,
	ioctl:		cmm_ioctl,
	open:		cmm_open,
	release:	cmm_close,
	/* NULL...*/
};

static dev_link_t *cardman_attach(void);	/*proto*/
static void cardman_detach(dev_link_t *link);	/*proto*/
static void cardman_release(u_long arg);	/*proto*/

static int major;	/* major number we get from the kernel */

/* note: the first state has to have number 0 always */

#define	M_FETCH_ATR	0
#define	M_TIMEOUT_WAIT	1
#define	M_READ_ATR_LEN	2
#define	M_READ_ATR	3
#define	M_ATR_PRESENT	4
#define	M_BAD_CARD	5
#define M_CARDOFF       6

#ifdef PCMCIA_DEBUG
char *m_stnames[]={
	"M_FETCH_ATR",
	"M_TIMEOUT_WAIT",
	"M_READ_ATR_LEN",
	"M_READ_ATR",
	"M_ATR_PRESENT",
	"M_BAD_CARD"
};
#endif

#define	LOCK_IO			0
#define	LOCK_MONITOR		1

#define IS_AUTOPPS_ACT          6
#define	IS_PROCBYTE_PRESENT	 7
#define	IS_INVREV		 8
#define IS_ANY_T0                9
#define	IS_ANY_T1		10
#define	IS_ATR_PRESENT		11
#define	IS_ATR_VALID		12
#define	IS_CMM_ABSENT           13
#define	IS_BAD_LENGTH		14
#define	IS_BAD_CSUM		15
#define	IS_BAD_CARD		16

typedef struct cardman_dev_t {
	dev_link_t	link;			/* pcmcia link */
	dev_node_t	node;			/* OS node (major,minor) */

	unsigned char	atr[MAX_ATR];
	unsigned char	rbuf[512];

	wait_queue_head_t devq;	/* when removing cardman
				 * must not be zeroed!
				 */

	wait_queue_head_t ioq;		/* if IO is locked, wait on this Q*/
	wait_queue_head_t atrq;		/* wait for ATR valid */
	wait_queue_head_t readq;	/* used by write to wake blk.read */

	/* warning - do not move this fields.
	 * initialising to zero depends on it -
	 * see ZERO_DEV below.
	 * these fields are zeroed upon close.
	 */
	unsigned char	atr_csum;
	unsigned char	atr_len_retry;
	unsigned short	atr_len;
	unsigned short	rlen;		/* bytes avail. after write */
	unsigned short	rpos;		/* latest read pos. write zeroes */
	unsigned char	procbyte;	/* T=0 procedure byte */
	unsigned char	mstate;		/* state of card monitor */
	unsigned char	cwarn;		/* slow down warning */
	unsigned char	flags0;		/* cardman IO-flags 0 */
	unsigned char	flags1;		/* cardman IO-flags 1 */
	unsigned int	mdelay;		/* variable monitor speeds */

	unsigned int	baudv;		/* baud value for speed */
        unsigned char   ta1;
	unsigned char 	proto;		/* T=0, T=1, ... */
	unsigned long	flags;		/* lock+flags (MONITOR,IO,ATR)
					 * for concurrent access
					 */
        unsigned char   pts[4];

	struct timer_list timer;	/* used to keep monitor running */
	struct task_struct *owner;	/* who opened us (only one per dev)*/
  //#if PCMCIA_DEBUG
	int monitor_running;
  //#endif
} cardman_dev_t;

#define	ZERO_DEV(dev)  \
	memset(&dev->atr_csum,0, \
		sizeof(cardman_dev_t) - \
		/*link*/ sizeof(dev_link_t) - \
		/*node*/ sizeof(dev_node_t) - \
		/*atr*/ MAX_ATR*sizeof(char) - \
		/*rbuf*/ 512*sizeof(char) - \
		/*queue*/ 4*sizeof(wait_queue_head_t));

static dev_info_t dev_info = MODULE_NAME;
static dev_link_t *dev_table[CM4000_MAX_DEV] = { NULL, };

static unsigned char ucFIDITable[10][14] =
{
//FI     00    01    02    03    04    05    06    07    08    09    10    11    12    13
//DI
/* 0 */{0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11},
/* 1 */{0x01, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x91, 0x11, 0x11, 0x11, 0x11},
/* 2 */{0x02, 0x12, 0x22, 0x32, 0x11, 0x11, 0x11, 0x11, 0x11, 0x92, 0xA2, 0xB2, 0x11, 0x11},
/* 3 */{0x03, 0x13, 0x23, 0x33, 0x43, 0x53, 0x63, 0x11, 0x11, 0x93, 0xA3, 0xB3, 0xC3, 0xD3},
/* 4 */{0x04, 0x14, 0x24, 0x34, 0x44, 0x54, 0x64, 0x11, 0x11, 0x94, 0xA4, 0xB4, 0xC4, 0xD4},
/* 5 */{0x00, 0x15, 0x25, 0x35, 0x45, 0x55, 0x65, 0x11, 0x11, 0x95, 0xA5, 0xB5, 0xC5, 0xD5},
/* 6 */{0x06, 0x16, 0x26, 0x36, 0x46, 0x56, 0x66, 0x11, 0x11, 0x96, 0xA6, 0xB6, 0xC6, 0xD6},
/* 7 */{0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11},
/* 8 */{0x08, 0x11, 0x28, 0x38, 0x48, 0x58, 0x68, 0x11, 0x11, 0x98, 0xA8, 0xB8, 0xC8, 0xD8},
/* 9 */{0x09, 0x19, 0x29, 0x39, 0x49, 0x59, 0x69, 0x11, 0x11, 0x99, 0xA9, 0xB9, 0xC9, 0xD9}
};

/*========================== code starts here ==========================*/

#ifndef PCMCIA_DEBUG
#define	xoutb	outb
#define	xinb	inb
#else
static inline void xoutb(unsigned char val,unsigned short port) {
  #if 1 
	DEBUG(6, "outb(val=%.2x,port=%.4x)\n",
		val, port);
	#endif
	outb(val,port);
}
static unsigned char xinb(unsigned short port) {
unsigned char val;
	val=inb(port);
#if 1
	DEBUG(6, "%.2x=inb(%.4x)\n",
		val,port);
#endif
	return val;
}

static void dumpreg(char *s,unsigned int iobase,int len) {
int i;
	DEBUG(5, "%s: ",s);
	for(i=0;i<len;i++)
		DEBUG(5, "%.2x ",inb(iobase+i));
	DEBUG(5, "\n");
}

static void dumpstat(int iobase) {
unsigned char b;
	b=inb(iobase);
	DEBUG(5, "f0=%.2x inserted=%s powered=%s txdone=%s\n",
		b,b&1?"yes":"no",b&2?"yes":"no",b&8?"yes":"no");
}

static void dev_dumpatr(void *p) {		/* debugging only */
register cardman_dev_t *dev=(cardman_dev_t *)p;
unsigned long flags;
int i;

	save_flags(flags);
	sti();
	DEBUG(5, "cardman_cs: dumpatr: ");
	if (test_bit(IS_ATR_VALID,&dev->flags)==0) {
		DEBUG(5, "no atr\n");
		return;
	}
	DEBUG(5, "atr_len=%d ",dev->atr_len);
	for(i=0;i<dev->atr_len;i++)
		DEBUG(5, "%.2x ",dev->atr[i]);
	DEBUG(5, "\n");
	restore_flags(flags);
}
#endif

#define	b_0000	15
#define	b_0001	14
#define	b_0010	13
#define	b_0011	12
#define	b_0100	11
#define	b_0101	10
#define	b_0110	9
#define	b_0111	8
#define	b_1000	7
#define	b_1001	6
#define	b_1010	5
#define	b_1011	4
#define	b_1100	3
#define	b_1101	2
#define	b_1110	1
#define	b_1111	0

static unsigned char irtab[16]={
	b_0000, b_1000, b_0100, b_1100,
	b_0010, b_1010, b_0110, b_1110,
	b_0001, b_1001, b_0101, b_1101,
	b_0011, b_1011, b_0111, b_1111
};

static void str_invert_revert(unsigned char *b,int len) {
int i;
	for(i=0;i<len;i++)
		b[i]=(irtab[b[i]&0x0f]<<4) | irtab[b[i]>>4];
}

static unsigned char invert_revert(unsigned char ch) {
	return (irtab[ch&0x0f]<<4) | irtab[ch>>4];
}

#define	ATRLENCK(dev,pos) \
	if (pos>=dev->atr_len || pos>=MAX_ATR) \
		goto return_0;

/* interruptible_pause() */

static inline void ipause(unsigned long amount) {
	current->state=TASK_INTERRUPTIBLE;
	schedule_timeout(amount);
}

/* uninterruptible_pause() */

static inline void upause(unsigned long amount) {
	current->state=TASK_UNINTERRUPTIBLE;
	schedule_timeout(amount);
}

static unsigned int calcBaudv(unsigned char fidi){
  unsigned int wcrcf, wbrcf, RFU; 

  RFU = 1;

  DEBUG(3, "cardman_cs: -> calcBaudv (fidi = 0x%.2x)\n", fidi);
  
  // FI
  switch ((fidi>>4)&0x0F){
  case 0x00: wcrcf = 372; break;
  case 0x01: wcrcf = 372; break;
  case 0x02: wcrcf = 558; break;
  case 0x03: wcrcf = 744; break;
  case 0x04: wcrcf = 1116; break;
  case 0x05: wcrcf = 1488; break;
  case 0x06: wcrcf = 1860; break;
  case 0x07: wcrcf = RFU; break;
  case 0x08: wcrcf = RFU; break;
  case 0x09: wcrcf = 512; break;
  case 0x0A: wcrcf = 768; break;
  case 0x0B: wcrcf = 1024; break;
  case 0x0C: wcrcf = 1536; break;
  case 0x0D: wcrcf = 2048; break;
  default: wcrcf = RFU;break;
  }
  DEBUG(5, "cardman_cs: calcBaudv: wcrcf = %i\n", wcrcf);

  // DI
  switch (fidi&0x0F){
  case 0x00: wbrcf = RFU; break;
  case 0x01: wbrcf = 1; break;
  case 0x02: wbrcf = 2; break;
  case 0x03: wbrcf = 4; break;
  case 0x04: wbrcf = 8; break;
  case 0x05: wbrcf = 16; break;
  case 0x06: wbrcf = 32; break;
  case 0x07: wbrcf = RFU; break;
  case 0x08: wbrcf = 12; break;
  case 0x09: wbrcf = 20; break; 
  default: wbrcf=RFU; break;
  }
  DEBUG(5, "cardman_cs: calcBaudv: wbrcf = %i\n", wbrcf);
  DEBUG(3, "cardman_cs: <- Calcbaudv returns 0x%.4x\n", (wcrcf/wbrcf));
  return (wcrcf/wbrcf);
}

static inline unsigned short IO_ReadNumRecBytes (int iobase,unsigned short *s) {
  /*unsigned short tmp;
  
  tmp=*s=0;
  do {
    tmp = inb(iobase+2)|(inb(iobase)&4?0x100:0);
  }while (tmp != *s);
  return *s;
  */
  return *s=inb(iobase+2)|(inb(iobase)&4?0x100:0);
  
}

static int parse_atr(cardman_dev_t *dev) {
unsigned char anyT1, anyT0;
unsigned char ch, ifno;
int ix, done;

         DEBUG(3, "cardman_cs: -> parse_atr: dev->atr_len = %i\n", dev->atr_len);

        if (dev->atr_len<3){
          DEBUG(5, "cardman_cs: parse_atr: atr_len < 3\n");
          goto return_0;
        }

	if (dev->atr[0]==0x3f) set_bit(IS_INVREV,&dev->flags);
	else clear_bit(IS_INVREV,&dev->flags);
	ix=1;
	ifno=1;
	ch=dev->atr[1];
	dev->proto=0;	/* XXX PROTO */
	anyT1=anyT0=done=0;
	dev->ta1=0x11; //defaults to 9600 baud*/
	do {
	        if (ifno==1 && (ch&0x10)){ // read first interface byte and TA1 is present
		  dev->ta1=dev->atr[2]; 
		  DEBUG(5, "cardman_cs: Card says FiDi is 0x%.2x\n", dev->ta1);
		  ifno++;
		}else if ((ifno==2) && (ch&0x10)){ //  TA(2)
		  dev->ta1 = 0x11;
		  ifno++;
		}

		DEBUG(5,"cardman_cs: Yi=%.2x\n",ch&0xf0);
		ix+=((ch&0x10)>>4)		/* no of int.face chars */
			+ ((ch&0x20)>>5)
			+ ((ch&0x40)>>6)
			+ ((ch&0x80)>>7);
		// ATRLENCK(dev,ix);
		if (ch&0x80) {	/* TDi */
			ch=dev->atr[ix];
			if ((ch&0x0f)) {
			  anyT1=1;
			  DEBUG(5, "cardman_cs: card is capable of T=1\n");
			}
			else{
			  anyT0=1;
			  DEBUG(5, "cardman_cs: card is capable of T=0\n");
			} 
		}
		else done=1;
	} while(!done);

#if 1
	DEBUG(5,"cardman_cs: ix=%d noHist=%d anyT1=%d\n",
		ix,dev->atr[1]&15,anyT1);
	if (ix+1+(dev->atr[1]&0x0f)+anyT1!=dev->atr_len) {
		DEBUG(5,"cardman_cs: length error\n");
		goto return_0;
	}
#endif
	if (anyT0)
	  set_bit(IS_ANY_T0, &dev->flags);
	if (anyT1) {	/* compute csum */
		dev->atr_csum=0;
#if 0
		for(i=1;i<dev->atr_len;i++)
			dev->atr_csum^=dev->atr[i];
		if (dev->atr_csum) {
			set_bit(IS_BAD_CSUM,&dev->flags);
			DEBUG(5,"cardman_cs: bad checksum\n");
			goto return_0;
		}
#endif
		if (anyT0==0) 
		  dev->proto=1;	/* XXX PROTO */
		set_bit(IS_ANY_T1,&dev->flags);
	}

	return 1;
return_0:
	return 0;
}

static void set_cardparameter(cardman_dev_t *dev) {
register int iobase=dev->link.io.BasePort1;
 DEBUG(3, "cardman_cs: -> set_cardparameter\n"); 
 //xoutb((0x20 | (((dev->baudv-1)&0x0100)>>8) ), iobase+1); // set Flags1
 xoutb(dev->flags1, iobase+1); //set flags1
 xoutb(dev->baudv ,iobase+6); // set baudrate
 // set stopbits
 if ((dev->atr[0] == 0x3B) &&
     (dev->atr[1] == 0xBE) &&
     (dev->atr[2] == 0x11) &&
     (dev->atr[3] == 0x00) &&
     (dev->atr[4] == 0x00) &&
     (dev->atr[5] == 0x41) &&
     (dev->atr[6] == 0x01) /* &&
     (dev->atr[7] == 0x38) &&
     (dev->atr[8] == 0x01) &&
     (dev->atr[9] == 0x00) &&
     (dev->atr[10] == 0x03) &&
     (dev->atr[11] == 0x00) &&
     (dev->atr[12] == 0x00) &&
     (dev->atr[13] == 0x00) &&
     (dev->atr[14] == 0x00) &&
     (dev->atr[15] == 0x00) &&
     (dev->atr[16] == 0x01) &&
     (dev->atr[17] == 0x90) &&
     (dev->atr[18] == 0x00)*/
    )
 { //ACOS
   DEBUG(5, "cardman_cs: Found a ACOS card - applying extra guard time\n");
   xoutb(0x03, iobase+7);
 }else if 
    ((dev->atr[0] == 0x3B) &&
     (dev->atr[1] == 0x76) &&
     (dev->atr[2] == 0x13) &&
     (dev->atr[3] == 0x00) &&
     (dev->atr[4] == 0x00) &&
     (dev->atr[5] == 0x80) &&
     (dev->atr[6] == 0x62) &&
     (dev->atr[7] == 0x07) &&
     (dev->atr[8] == 0x41) &&
     (dev->atr[9] == 0x81) &&
     (dev->atr[10] == 0x81)
    )
   {//Motorola
   xoutb(0x04, iobase+7);
   }else{ // ISO default
   xoutb(0x02, iobase+7);
 }

 DEBUG(3, "cardman_cs: <- set_cardparameter\n");
}

static int set_protocol(cardman_dev_t *dev, ptsreq_t *ptsreq){
 
  unsigned long tmp, i;
  unsigned short numBytesRead;
  unsigned char ptsReply[4];
  ssize_t rc;
  register int iobase=dev->link.io.BasePort1;

  rc = 0;

  DEBUG(3, "cardman_cs: -> set_protocol\n");
  DEBUG(5, "cardman_cs: ptsreq->Protocol = 0x%.8x\nptsreq->Flags=0x%.2x\nptsreq->pts1=0x%.2x\nptsreq->pts2=0x%.2x\nptsreq->pts3=0x%.2x\n", 
	ptsreq->protocol, ptsreq->flags, ptsreq->pts1, ptsreq->pts2, ptsreq->pts3);

  // Fill PTS structure
  dev->pts[0] = 0xff;
  dev->pts[1] = 0x00;
  tmp = ptsreq->protocol;
  while((tmp=(tmp>>1))>0){dev->pts[1]++;}
  dev->proto=dev->pts[1]; // Set new protocol
  dev->pts[1]=(0x01<<4)|(dev->pts[1]);
  
  // Correct Fi/Di according to CM4000 Fi/Di table
  DEBUG(5, "cardman_cs: Ta(1) from ATR is 0x%.2x\n", dev->ta1);
  dev->pts[2] = ucFIDITable[dev->ta1&0x0F][(dev->ta1>>4)&0x0F]; //set Fi/Di according to ATR TA(1)
  
  // Calculate PCK character
  dev->pts[3]=dev->pts[0]^dev->pts[1]^dev->pts[2];

  DEBUG(5, "cardman_cs: pts0 = %.2x\n", dev->pts[0]);
  DEBUG(5, "cardman_cs: pts1 = %.2x\n", dev->pts[1]);
  DEBUG(5, "cardman_cs: pts2 = %.2x\n", dev->pts[2]);
  DEBUG(5, "cardman_cs: pts3 = %.2x\n", dev->pts[3]);
  
  //  if (dev->pts[2] != 0x11){
    //check card convention
    if (test_bit(IS_INVREV, &dev->flags)){
      str_invert_revert(dev->pts, 4);
    }

    //reset SM
    xoutb(0x80, iobase);

    // Enable access to the message buffer
    DEBUG(5, "cardman_cs: Enable access to the messages buffer\n");
    dev->flags1=0x20		/* T_Active */
      | (test_bit(IS_INVREV,&dev->flags)?0x02:0x00) /* inverse parity */
      //| (dev->baudv&0x100?1:0);	/* MSB-Baud */
      | ((dev->baudv>>8)&0x01); /* MSB-baud */
    xoutb(dev->flags1,
	  iobase+1);

    DEBUG(5, "cardman_cs: Enable message buffer -> flags1 = 0x%.2x\n", dev->flags1);

    // write challenge to the buffer
    DEBUG(5, "cardman_cs: Write challenge to buffer: ");
    for(i=0;i<4;i++) {
      xoutb(i,iobase+3);	/* buf addr */
      xoutb(dev->pts[i],iobase+4); /* buf data */
      DEBUG(5, "0x%.2x ",dev->pts[i]);
    }
    DEBUG(5, "\n");
    
    // set number of bytes to write
    DEBUG(5, "cardman_cs: Set number of bytes to write\n");
    xoutb(0x04,iobase+5);

    // Trigger CARDMAN CONTROLLER
    xoutb(0x50, iobase);

    
    // Monitor progress
    /* wait for xmit done */
    DEBUG(5, "cardman_cs: Waiting for NumRecBytes getting valid\n");
    
    for(i=0;i<100;i++) {
      if (inb(iobase)&0x08){
	DEBUG(5, "cardman_cs: NumRecBytes is valid\n");
	break;
      }
      mdelay(T_10MSEC);
    }
    if (i==100) {
      DEBUG(5, "cardman_cs: Timeout waiting for NumRecBytes getting valid\n");
      rc=-EIO;
      goto exit_setprotocol;
    }

    DEBUG(5, "cardman_cs: Reading NumRecBytes\n");
    for(i=0;i<100;i++) { 
      IO_ReadNumRecBytes(iobase,&numBytesRead);
      if (numBytesRead >= 4){
	DEBUG(1, "cardman_cs:  NumRecBytes = %i\n", numBytesRead);
	break;
      }
      mdelay(T_10MSEC);
    }

    if (numBytesRead == 3) // short PTS reply!?
      i=0;

    if (i==100){
      DEBUG(5, "cardman_cs: Timeout reading numBytesRead\n");
      rc=-EIO;
      goto exit_setprotocol;
    }    
    
    // Reset the CARDMAN CONTROLLER
    DEBUG(5, "cardman_cs: Reset the CARDMAN CONTROLLER\n");
    xoutb(0x80,iobase);

    // Read PPS reply
    DEBUG(5, "cardman_cs: Read PPS reply\n");
    for(i=0;i<numBytesRead;i++) {
      xoutb(i,iobase+3);	// buf addr
      ptsReply[i]=inb(iobase+4);   // buf data 
    }

#ifdef PCMCIA_DEBUG
    DEBUG(1, "cardman_cs: PTSreply: ");
    for (i=0; i<numBytesRead; i++){
      DEBUG(1, "0x%.2x ", ptsReply[i]);
    }
    DEBUG(1, "\n");
#endif // PCMCIA_DEBUG
    
    DEBUG(5, "cardman_cs: Clear Tactive in Flags1\n");
    xoutb(0x20,iobase+1);	// Clear Tactive in Flags1

    // Compare ptsreq and ptsreply
    if ((dev->pts[0] == ptsReply[0]) &&
        (dev->pts[1] == ptsReply[1]) &&
	(dev->pts[2] == ptsReply[2]) &&
	(dev->pts[3] == ptsReply[3])){
      // setcardparameter according to PPS
      dev->baudv=calcBaudv(dev->pts[2]);
      set_cardparameter(dev);
    }else if ((dev->pts[0] == ptsReply[0]) &&
	      ((dev->pts[1]&0xef) == ptsReply[1]) &&
	      ((ptsReply[0]^ptsReply[1]) == ptsReply[2])){ //short PTS reply
      //set card parameter to default values
      dev->baudv = calcBaudv(0x11);
      set_cardparameter(dev);
    }else{
      rc=-EIO;
    }
    /*
  }// else: default values apply so no PPS necessary 
  else{
    DEBUG(5, "FIDI = 0x11 -> no PPS necessary <skip>\n");
    set_cardparameter(dev); //in case its a card which needs extra guard time 
    rc=0;
  }
*/
exit_setprotocol:
  DEBUG(3, "cardman_cs: <- set_protocol\n");
  return rc;
}

static inline int IO_DetectCMM(int iobase,cardman_dev_t *dev) {

	/* note: statemachine is assumed to be reset */
	if (inb(iobase)&8) {
		clear_bit(IS_ATR_VALID,&dev->flags);
		set_bit(IS_CMM_ABSENT,&dev->flags);
		return 0;		/* detect CMM = 1 -> failure */
	}
	//xoutb(0x40,iobase+1);		/* detectCMM */
	xoutb(dev->flags1|0x40, iobase+1);
	if ((inb(iobase)&8)==0) {
		clear_bit(IS_ATR_VALID,&dev->flags);
		set_bit(IS_CMM_ABSENT,&dev->flags);
		return 0;		/* detect CMM=0 -> failure */
	}
	/* clear detectCMM again by restoring original flags1 */
	xoutb(dev->flags1,iobase+1);
	return 1;
}

static void terminate_monitor(cardman_dev_t *dev) {

	/* tell the monitor to stop and wait until
	 * it terminates.
	 */
        DEBUG(3, "cardman_cs: -> terminate_monitor\n");
	while(test_and_set_bit(LOCK_MONITOR,&dev->flags)) {
		sleep_on(&dev->devq);
	}
	/* now, LOCK_MONITOR has been set.
	 * allow a last cycle in the monitor.
	 * the monitor will indicate that it has
	 * finished by clearing this bit.
	 */
	DEBUG(5, "cardman_cs: Now allow last cycle of monitor!\n");
	while(test_bit(LOCK_MONITOR,&dev->flags)) {
		upause(T_25MSEC);
	}
	DEBUG(5, "cardman_cs: Delete timer\n");
	del_timer(&dev->timer);
#ifdef PCMCIA_DEBUG
	dev->monitor_running=0;
#endif

	 DEBUG(3, "cardman_cs: <- terminate_monitor\n");
}

/*
 * monitor the card every 50msec. as a side-effect, retrieve the
 * atr once a card is inserted. another side-effect of retrieving the
 * atr is that the card will be powered on, so there is no need to
 * power on the card explictely from the application: the driver
 * is already doing that for you.
 */

static void monitor_card(unsigned long p) {
register cardman_dev_t *dev=(cardman_dev_t *)p;
register int iobase=dev->link.io.BasePort1;
unsigned short s;
ptsreq_t ptsreq;
int i, atrc;

        DEBUG(2, "cardman_cs: ->  monitor_card\n"); 
 
	/* if someone has set the lock for us: we're done!*/
	if (test_and_set_bit(LOCK_MONITOR,&dev->flags)) {
	  DEBUG(4, "cardman_cs: About to stop monitor\n");
	  /* no */
	  dev->rlen=
	  dev->rpos=
	  dev->atr_csum=
	  dev->atr_len_retry=
	  dev->cwarn=
	  dev->mstate=0;
	  clear_bit(LOCK_MONITOR,&dev->flags);
	  wake_up(&dev->devq);	/* close et al. are sleeping on it */
	  DEBUG(1, "cardman_cs: <- monitor_card (we are done now)\n");
	  return;
	}

	/* try to lock io: if it is already locked, just add another timer */
	if (test_and_set_bit(LOCK_IO,(void *)&dev->flags)){
	        DEBUG(4, "cardman_cs: monitor_card: Couldn't get IO lock\n");
	        goto return_with_timer;
	}

	/* is a card/a reader inserted at all ? */
	dev->flags0=xinb(iobase);
	DEBUG(1,"cardman_cs: monitor_card:  dev->flags0 = 0x%2x\n", dev->flags0);
	DEBUG(1,"cardman_cs: monitor_card:  smartcard present: %s\n", dev->flags0&1?"yes":"no");
	DEBUG(1,"cardman_cs: monitor_card:  cardman present: %s\n", dev->flags0==0xff?"no":"yes");
	  
	if ((dev->flags0&1)==0		/* no smartcard inserted */
		|| dev->flags0==0xff)	/* no cardman inserted */
	{
		/* no */
		dev->rlen=
		dev->rpos=
		dev->atr_csum=
		dev->atr_len_retry=
		dev->cwarn=
		dev->mstate=0;

		dev->flags&=0x000000ff; // only keep IO and MONITOR locks
		
		if (dev->flags0==0xff){
		  DEBUG(4, "cardman_cs: monitor_card: set IS_CMM_ABSENT bit\n"); 
		  set_bit(IS_CMM_ABSENT,&dev->flags);
		}else if (test_bit(IS_CMM_ABSENT, &dev->flags)){
		  DEBUG(4, "cardman_cs: monitor_card: clear IS_CMM_ABSENT bit (card is removed)\n"); 
		  clear_bit(IS_CMM_ABSENT,&dev->flags);
		}
		DEBUG(4, "cardman_cs: monitor_card: dev->flags = 0x%.8x\n", (unsigned long)dev->flags);

		goto release_io;
	}else if ((dev->flags0&1) && (test_bit(IS_CMM_ABSENT, &dev->flags))) { 
	        // cardman and card present but cardman was absent before (after suspend with inserted card)
	        DEBUG(4, "cardman_cs: monitor_card: clear IS_CMM_ABSENT bit (card is inserted)\n");
	        clear_bit(IS_CMM_ABSENT, &dev->flags);
	}

	if (test_bit(IS_ATR_VALID,&dev->flags)==1){
		DEBUG(4, "cardman_cs: monitor_card believes ATR is already valid (do nothing)\n");
		goto release_io;
	}
  
	switch(dev->mstate) {
	case M_CARDOFF:
	        { unsigned char flags0;
		DEBUG(4, "cardman_cs: monitor_card: M_CARDOFF\n");
	          flags0 = inb(iobase);
		  if (flags0 & 0x02){ // wait until Flags0 indicate power is off
		    dev->mdelay=T_10MSEC;
		  }else{		  
		    // Flags0 indicate power off and no card inserted now; Reset CARDMAN CONTROLLER
		    xoutb(0x80, iobase);
		   		    
		    // prepare for fetching ATR again: after card off ATR is read again automatically
		    dev->rlen=
		    dev->rpos=
		    dev->atr_csum=
		    dev->atr_len_retry=
		    dev->cwarn=
		    dev->mstate=0;
 		   
		    // minimal gap between CARDOFF and read ATR is 50 msec
		    dev->mdelay=T_50MSEC;	        
		  }
		}
	        break;
	case M_FETCH_ATR:
	        DEBUG(4, "cardman_cs: monitor_card: M_FETCH_ATR\n");
		xoutb(0x80,iobase);
		DEBUG(4, "cardman_cs: Reset BAUDV to 9600\n");
		dev->baudv=0x173;	/* 9600 */
		xoutb(0x02,iobase+7);	/* stopbits=2 */
		xoutb(0x73,iobase+6);	/* baud value */
		xoutb(0x21,iobase+1);	/* T_Active=1, baud value */
		xoutb(dev->flags0&2?0x46:0x44,iobase);	/* warm start vs. power on */
		dev->mdelay=T_40MSEC;
		dev->mstate=M_TIMEOUT_WAIT;
		break;
	case M_TIMEOUT_WAIT:
	        DEBUG(4, "cardman_cs: monitor_card:  M_TIMEOUT_WAIT\n");
		/* numRecBytes */
		IO_ReadNumRecBytes(iobase,&dev->atr_len);
		dev->mdelay=T_10MSEC;
		dev->mstate=M_READ_ATR_LEN;
		break;
	case M_READ_ATR_LEN:
	        DEBUG(4, "cardman_cs: monitor_card: M_READ_ATR_LEN\n");
		/* endlosschleife evtl. moeglich, da kein timeout */
		// dumpreg("monitor",iobase,9);

#define	MAX_ATR_LEN_RETRY	100

		if (dev->atr_len==IO_ReadNumRecBytes(iobase,&s)) {
		  if (dev->atr_len_retry++>=MAX_ATR_LEN_RETRY) { /* + XX msec */
		    dev->mdelay=T_10MSEC;//0;
		    dev->mstate=M_READ_ATR;
		  }
		}
		else {
		  dev->atr_len=s;
		  dev->atr_len_retry=0;	/* set new timeout */
		}

#if 0
		if ((dev->atr_len<3 && dev->atr_len_retry>=MAX_ATR_LEN_RETRY)
			|| dev->atr_len>MAX_ATR) {
			set_bit(IS_BAD_LENGTH,&dev->flags);
			dev->mdelay=0;
			dev->mstate=M_BAD_CARD;
		}
#endif
		DEBUG(4, "cardman_cs: monitor_card: Current ATR_LEN = %i\n", dev->atr_len);
		break;
	case M_READ_ATR:
	        DEBUG(4, "cardman_cs: monitor_card: M_READ_ATR\n");
		xoutb(0x80,iobase);	/* reset SM */
		for(i=0;i<dev->atr_len;i++) {
			xoutb(i,iobase+3);		/* buf addr */
			dev->atr[i]=inb(iobase+4);	/* buf data */
		}
		// Deactivate T_Active flags
		DEBUG(4, "cardman_cs: monitor_card: Deactivate T_Active flags\n");
		dev->flags1 = 0x01;
		xoutb(dev->flags1, iobase+1);
		
		/* atr is present (which doesnt mean it's valid) */
		set_bit(IS_ATR_PRESENT,&dev->flags);
		if (dev->atr[0]==0x03)
			str_invert_revert(dev->atr,dev->atr_len);
		atrc=parse_atr(dev);
		if (atrc==0) {	/* atr invalid */
		  dev->mdelay=0;
		  dev->mstate=M_BAD_CARD;
		}
		else {
		  dev->mdelay=T_50MSEC;
		  dev->mstate=M_ATR_PRESENT;
		  set_bit(IS_ATR_VALID,&dev->flags);
		}
				
		if (test_bit(IS_ATR_VALID, &dev->flags) == 1){
		  DEBUG(4, "cardman_cs: monitor_card: ATR valid\n");
		  if( (test_bit(IS_AUTOPPS_ACT, &dev->flags)==0) &&
                      (dev->ta1 != 0x11) &&  //nop PPS necessary (default values)
                      !(test_bit(IS_ANY_T0, &dev->flags) && test_bit(IS_ANY_T1, &dev->flags))) //do not do PPS with multi protocol cards
                  {
		    DEBUG(4, "cardman_cs: monitor_card: Perform AUTOPPS\n");
		    set_bit(IS_AUTOPPS_ACT, &dev->flags);
		    ptsreq.protocol = ptsreq.protocol = (0x01<<dev->proto);
		    ptsreq.flags = 0x01;
		    ptsreq.pts1 = 0x00;
		    ptsreq.pts2 = 0x00;
		    ptsreq.pts3 = 0x00;
		    if (set_protocol(dev, &ptsreq) == 0){
		      DEBUG(4, "cardman_cs: monitor_card: AUTOPPS returns SUCC\n");
		      clear_bit(IS_AUTOPPS_ACT, &dev->flags);
		      wake_up_interruptible(&dev->atrq);
		    }else{
		      DEBUG(4, "cardman_cs: monitor_card: AUTOPPS failed: repower using defaults\n");
		      // prepare for repowering 
		      clear_bit(IS_ATR_PRESENT, &dev->flags);
		      clear_bit(IS_ATR_VALID, &dev->flags);
		      dev->rlen=
			dev->rpos=
			dev->atr_csum=
			dev->atr_len_retry=
			dev->cwarn=
			dev->mstate=0;
	 	      
		        dev->mdelay=T_50MSEC;
		      }
		  }else{
		    set_cardparameter(dev); // for cards which use slightly different params (extra guard time)
		    if (test_bit(IS_AUTOPPS_ACT, &dev->flags) == 1) DEBUG(4, "cardman_cs: monitor_card: AUTOPPS already active  (2nd try - use default values)\n");
		    if (dev->ta1 == 0x11) DEBUG(4, "cardman_cs: monitor_card: No AUTOPPS necessary TA(1)==0x11\n");
		    if (test_bit(IS_ANY_T0, &dev->flags) && test_bit(IS_ANY_T1, &dev->flags)) DEBUG(4, "cardman_cs: monitor_card: Do NOT perform AUTOPPS with multiprotocol cards\n");
		    clear_bit(IS_AUTOPPS_ACT, &dev->flags);		    
		    wake_up_interruptible(&dev->atrq);
		  }
		}else{
		  DEBUG(4, "cardman_cs: monitor_card: ATR invalid\n");
		  wake_up_interruptible(&dev->atrq);
		}
		break;
	case M_BAD_CARD:
	        DEBUG(4, "cardman_cs: monitor_card: M_BAD_CARD\n");
		/* slow down warning, but prompt immediately
		 * after insertion
		 */
		if (dev->cwarn==0 || dev->cwarn==10) {
			set_bit(IS_BAD_CARD,&dev->flags);
			printk(KERN_WARNING MODULE_NAME ": device %s: ",
				dev->node.dev_name);
			if (test_bit(IS_BAD_CSUM,&dev->flags)) {
			  DEBUG(4, "ATR checksum (0x%.2x, should be zero) "
				"failed\n",
			  dev->atr_csum);
			}
			#ifdef PCMCIA_DEBUG
			else if (test_bit(IS_BAD_LENGTH,&dev->flags)){
			  DEBUG(4, "ATR length error\n");
			}else{ 
			  DEBUG(4, "card damaged or wrong way inserted\n");
			}
			#endif //PCMCIA_DEBUG
			dev->cwarn=0;
			wake_up_interruptible(&dev->atrq);	/* wake open */
		}
		dev->cwarn++;
		dev->mdelay=T_100MSEC;
		dev->mstate=M_FETCH_ATR;
		break;
	default:
	        DEBUG(4, "cardman_cs: monitor_card: Unknown action\n");
		break;	/* nothing */
	}

release_io:
	DEBUG(4, "cardman_cs: monitor_card: release_io\n");
	clear_bit(LOCK_IO,&dev->flags);
	wake_up_interruptible(&dev->ioq);	/* whoever needs IO */

return_with_timer:
	DEBUG(2,"cardman_cs: <- monitor_card (returns with timer)\n");
	dev->timer.expires=jiffies+dev->mdelay;
	add_timer(&dev->timer);
	clear_bit(LOCK_MONITOR,&dev->flags);
}

/*==== Interface to userland (file_operations) =========================*/

static loff_t cmm_llseek(struct file * filp, loff_t off, int whence) {
	DEBUG(3,"cardman_cs: cmm_llseek()\n");
	//	MAJOR(inode->i_rdev),MINOR(inode->i_rdev));
	return -ESPIPE;
}

static ssize_t cmm_read(struct file *filp,char *buf,size_t count,loff_t *ppos) {
register cardman_dev_t *dev=(cardman_dev_t *)filp->private_data;
register int iobase=dev->link.io.BasePort1;
unsigned long flags;
ssize_t rc;
int i,j,k;

	DEBUG(2, "cardman_cs: -> cmm_read(%s,%d)\n",
		current->comm,current->pid);

	if (count==0)		/* according to manpage */
		return 0;

	if ((dev->link.state & DEV_PRESENT)==0 ||	/* socket removed */
		test_bit(IS_CMM_ABSENT,&dev->flags))
		return -ENODEV;

	if (test_bit(IS_BAD_CSUM,&dev->flags))
		return -EIO;

	/* also see the note about this in cmm_write */

	save_flags(flags);
	sti();
	while(test_bit(IS_ATR_PRESENT,&dev->flags)==0) {
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		interruptible_sleep_on(&dev->atrq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}
	restore_flags(flags);

	if (test_bit(IS_ATR_VALID,&dev->flags)==0)
		return -EIO;

	/* this one implements blocking IO */

	while(dev->rpos>=dev->rlen) {
		if (filp->f_flags & O_NONBLOCK)
			return -EAGAIN;
		interruptible_sleep_on(&dev->readq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	/* lock io */
	while(test_and_set_bit(LOCK_IO,&dev->flags)) {
		interruptible_sleep_on(&dev->ioq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	rc=0;
	dev->flags0=inb(iobase);
	if ((dev->flags0&1)==0		/* no smartcard inserted */
		|| dev->flags0==0xff) {	/* no cardman inserted */
		clear_bit(IS_ATR_VALID,&dev->flags);
		if (dev->flags0&1) {
			set_bit(IS_CMM_ABSENT,&dev->flags);
			rc=-ENODEV;
		}
		rc=-EIO;
		goto release_io;
	}

#if 0
	if ((dev->proto && dev->rpos==0) ||	/* first read after write */
	    (dev->proto==0 && dev->rpos==5)) {
#endif
	  /* DEBUG(0, "cardman_cs: Read reply: ");
	  for (i=0; i<dev->rlen; i++){
	    xoutb(i,iobase+3);	// buf addr 
	    ch=xinb(iobase+4);   // buf data 
	    DEBUG(0, " 0x%.2x",  ch);
	  }
	    DEBUG(0, "\n");
	  */
	    
	  DEBUG(4, "cardman_cs: begin read answer\n");
		j=min(count,dev->rlen-dev->rpos);
		k=dev->rpos;
		if (k+j>255) j=256-k;
		DEBUG(4,"cardman_cs: read1 j=%d\n",j);
		/* copy directly to user space */
		for(i=0;i<j;i++) {
			xoutb(k++,iobase+3);	/* buf addr */
			buf[i]=xinb(iobase+4);   /* buf data */
		}
		j=min(count,dev->rlen-dev->rpos);
		if (k+j>255) {
			DEBUG(4, "cardman_cs: read2 j=%d\n",j);
			dev->flags1|=0x10;	/* MSB buf addr set */
			xoutb(dev->flags1,iobase+1);
			for(;i<j;i++) {
				xoutb(k++,iobase+3);	/* buf addr */
				buf[i]=xinb(iobase+4);   /* buf data */
			}
		}
#if 0
		DEBUG(4,"cardman_cs: rpos=0, copy remainder=%d\n",dev->rlen-j);
		/* copy remainder to device buffer */
		for (;i<dev->rlen;i++) {
			xoutb(i,iobase+3);	/* buf addr */
			dev->rbuf[i]=xinb(iobase+4); /* buf data */
		}
#endif
	if (dev->proto==0 && count>dev->rlen-dev->rpos) {
	  DEBUG(4, "cardman_cs: T=0 and count > buffer\n");
	  buf[i]=buf[i-1];
	  buf[i-1]=dev->procbyte;
	  j++;
	}
	count=j;
#if 0
	}
	else {	/* read from previously filled buffer */
		count=min(count,dev->rlen);
		DEBUG(4,"cardman_cs: read: copy from buf, rpos=%d count=%d\n",
			dev->rpos,count);
		copy_to_user(buf, dev->rbuf+dev->rpos, count);
	}
#endif

	dev->rpos=dev->rlen+1;

	// Clear T1Active
	DEBUG(4, "cardman_cs: Clear T1Active\n");
	dev->flags1 &= 0xdf;
	xoutb(dev->flags1,
		iobase+1);

	xoutb(0,iobase+1);	/* clear detectCMM */
	/* last check before exit */
	if (!IO_DetectCMM(iobase,dev))
		count=-ENODEV;

	if (test_bit(IS_INVREV,&dev->flags) && count>0)
		str_invert_revert(buf,count);

release_io:
	clear_bit(LOCK_IO,&dev->flags);
	wake_up_interruptible(&dev->ioq);

	DEBUG(2,"cardman_cs: <- cmm_read returns: rc = %i\n",  (rc<0?rc:count));
	return rc<0?rc:count;
}

static ssize_t cmm_write(struct file *filp,const char *buf,size_t count,
				loff_t *ppos) {
register cardman_dev_t *dev=(cardman_dev_t *)filp->private_data;
register int iobase=dev->link.io.BasePort1;
unsigned long flags;
unsigned short s;
unsigned char tmp;
unsigned char infolen;
unsigned char sendT0;
unsigned short nsend;
unsigned short nr;
ssize_t rc;
int i;

	DEBUG(2, "cardman_cs: -> cmm_write(%s,%d)\n",
		current->comm,current->pid);

	if (count==0)		/* according to manpage */
		return 0;

	if (dev->proto==0 && count<4) {	/* T0 must have at least 4 bytes */
		DEBUG(4,"cardman_cs: T0 short write\n");
		return -EIO;
	}

	nr=count&0x1ff;	/* max bytes to write */

	sendT0=dev->proto?0:nr>5?0x08:0;

	if ((dev->link.state & DEV_PRESENT)==0 ||	/* socket removed */
		test_bit(IS_CMM_ABSENT,&dev->flags))
		return -ENODEV;

	if (test_bit(IS_BAD_CSUM,&dev->flags)) {
		DEBUG(4,"cardman_cs: bad csum\n");
		return -EIO;
	}

	/*
	 * wait for atr to become valid.
	 * note: it is important to lock this
	 * code. if we dont, the monitor could
	 * be run between test_bit and the
	 * the call the sleep on the atr-queue.
	 * if *then* the monitor detects atr valid,
	 * it will wake up any process on the atr-queue,
	 * *but* since we have been interrupted, we do
	 * not yet sleep on this queue. this would result
	 * in a missed wake_up and the calling process
	 * would sleep forever (until interrupted).
	 * also, do *not* restore_flags before sleep_on,
	 * because this could result in the same situation!
	 */

	save_flags(flags);
	sti();
	while(test_bit(IS_ATR_PRESENT,&dev->flags)==0) {	/* no atr */
		if (filp->f_flags & O_NONBLOCK) {
			// restore_flags(flags);
			return -EAGAIN;
		}
		interruptible_sleep_on(&dev->atrq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}
	restore_flags(flags);

	if (test_bit(IS_ATR_VALID,&dev->flags)==0) {	/* invalid atr */
		DEBUG(4, "cardman_cs: invalid ATR\n");
		return -EIO;
	}

	/* lock io */
	while(test_and_set_bit(LOCK_IO,&dev->flags)) {
		interruptible_sleep_on(&dev->ioq);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	rc=0;
	dev->flags0=inb(iobase);
	if ((dev->flags0&1)==0	  /* no smartcard inserted */
	    || dev->flags0==0xff) /* no cardman inserted */
        {
	  clear_bit(IS_ATR_VALID,&dev->flags);
	  if (dev->flags0&1) {
	    set_bit(IS_CMM_ABSENT,&dev->flags);
	    rc=-ENODEV;
	  }
	  else {
	    DEBUG(4,"cardman_cs: IO error\n");
	    rc=-EIO;
	  }
	  goto release_io;
	}
	
	xoutb(0x80,iobase);		/* reset SM */

	if (!IO_DetectCMM(iobase,dev)){
	  rc=-ENODEV;
	  goto release_io;
	}

	dev->flags1 |= (sendT0); // reflect T=0 send/read mode in flags1

	set_cardparameter(dev);

	tmp=inb(iobase+1); //dummy read to reset flag procedure received

	dev->flags1=0x20			/* T_Active */
	           | (sendT0)
	           | (test_bit(IS_INVREV,&dev->flags)?2:0) /* inverse parity */
	           | (((dev->baudv-1)&0x0100)>>8);   /* MSB-Baud */
	DEBUG(4, "cardman_cs: set dev->flags1 = 0x%.2x\n", dev->flags1);
	xoutb(dev->flags1,
		iobase+1);

	/* xmit data */
	DEBUG(4, "cardman_cs: Xmit data\n");
	for(i=0;i<nr;i++) {
	  if (i>=256){
	    dev->flags1=0x20			        /* T_Active */
	                | (sendT0)			/* SendT0 */
		        | (test_bit(IS_INVREV,&dev->flags)?2:0) /* inverse parity */
		        | (((dev->baudv-1)&0x0100)>>8)   /* MSB-Baud */
		        | 0x10; /* set address high  */ 
	    DEBUG(4, "cardman_cs: dev->flags = 0x%.2x - set address high\n", dev->flags1);
	    xoutb(dev->flags1,
		  iobase+1);
	  }
	  if (test_bit(IS_INVREV,&dev->flags)) {
	    DEBUG(4, "cardman_cs: Apply inverse convention for 0x%.2x -> 0x%.2x\n", (unsigned char) buf[i], invert_revert(buf[i]));
	    xoutb(i,iobase+3);	/* buf addr */
	    xoutb(invert_revert(buf[i]),iobase+4); /* buf data */
	  } else {
	    xoutb(i,iobase+3);	/* buf addr */
	    xoutb(buf[i],iobase+4);	/* buf data */
	  }
	}
	DEBUG(4, "cardman_cs: Xmit done\n");

	if (dev->proto==0) {
		/* T=0 proto: 0 byte reply */
		if (nr==4) {
		        DEBUG(4, "cardman_cs: T=0 assumes 0 byte reply\n");
			xoutb(i,iobase+3);
			if (test_bit(IS_INVREV,&dev->flags))
				xoutb(0xff,iobase+4);
			else	xoutb(0x00,iobase+4);
		}
		
		/* numSendBytes */
		if (sendT0) nsend=nr;
		else {
		  if (nr==4) nsend=5;
		  else {
		    nsend=5+(unsigned char)buf[4];
		    if (buf[4]==0) nsend+=0x100;
		  }
		}
	}
	else nsend=nr;
	
	/* T0: output procedure byte */
	if (test_bit(IS_INVREV,&dev->flags)){
	  DEBUG(4, "cardman_cs: T=0 set Procedure byte (inverse-reverse) 0x%.2x \n", invert_revert(buf[1]));
	  xoutb(invert_revert(buf[1]),iobase+2);
	}else{
	  DEBUG(4, "cardman_cs: T=0 set Procedure byte 0x%.2x\n", buf[1]);
	  xoutb(buf[1],iobase+2);
	}

	//nsend=dev->proto?nr:sendT0?nr:nr==4?5:(unsigned char)buf[4]+5; /*protohdr+5*/
	DEBUG(4, "cardman_cs: set NumSendBytes = 0x%.2x\n", (unsigned char)(nsend&0xff)); 
	xoutb((unsigned char)(nsend&0xff),iobase+5);
	        
	DEBUG(4, "cardman_cs: Trigger CARDMAN CONTROLLER (0x%.2x)\n",
	      0x40				/* SM_Active */
	      | (dev->flags0&2?0:4)		/* power on if needed */
	      | (dev->proto?0x10:0x08)	/* T=1/T=0 */
	      | (nsend&0x100)>>8		/* MSB numSendBytes */ );
	xoutb(0x40				/* SM_Active */
		| (dev->flags0&2?0:4)		/* power on if needed */
		| (dev->proto?0x10:0x08)	/* T=1/T=0 */
		| (nsend&0x100)>>8,		/* MSB numSendBytes */
		iobase);

	/* wait for xmit done */
	if (dev->proto == 1){
	  DEBUG(4, "cardman_cs: Wait for xmit done\n");
	  for(i=0;i<1000;i++) {
	    if (inb(iobase)&0x08)
	      break;
	    ipause(T_10MSEC);
	  }
	  if (i==1000) {
	    DEBUG(4, "cardman_cs: timeout waiting for xmit done\n");
	    rc=-EIO;
	    goto release_io;
	  }
	}

	/* T=1: wait for infoLen */

	infolen=0;
	if (dev->proto) {
		/* wait until infoLen is valid */
		for(i=0;i<6000;i++) {	/* max waiting time of 1 min */
			IO_ReadNumRecBytes(iobase,&s);
			if (s>=3) {
				infolen=inb(iobase+1);
				DEBUG(4, "cardman_cs: infolen=%d\n",infolen);
				break;
			}
			ipause(T_10MSEC);
		}
		if (i==6000) {
			DEBUG(4, "cardman_cs: timeout waiting for infoLen\n");
			rc=-EIO;
			goto release_io;
		}
	}
	else clear_bit(IS_PROCBYTE_PRESENT,&dev->flags);

	/* numRecBytes | bit9 of numRecytes */
	IO_ReadNumRecBytes(iobase,&dev->rlen);
	for(i=0;i<200;i++) {	/* max waiting time of 2 sec */
		if (dev->proto) {
		  if (dev->rlen>=infolen+4)
		    break;
		}
		ipause(T_10MSEC);
		/* numRecBytes | bit9 of numRecytes */
		IO_ReadNumRecBytes(iobase,&s);
		if (s>dev->rlen) {
		  DEBUG(4, "cardman_cs: NumRecBytes incremented (reset timeout)\n");
		  i=0;	/* reset timeout */
		  dev->rlen=s;
		}
		/* T=0: we are done when numRecBytes doesn't
		 *      increment any more and NoProcedureByte
		 *	is set and numRecBytes == bytes sent + 6
		 *	(header bytes + data + 1 for sw2)
		 *      except when the card replies an error
		 *      which means, no data will be sent back.
		 */
		else 
		if (dev->proto==0) {
		  if ((inb(iobase+3)&0x80)==0x80) { // no procedure byte received since last read
		    DEBUG(4, "cardman_cs: NoProcedure byte set\n");
		    //i=0;
		  }else{ // procedure byte received since last read
		    DEBUG(4, "cardman_cs:  NoProcedure byte unset (reset timeout)\n");
		    dev->procbyte=inb(iobase+1);
		    DEBUG(4, "cardman_cs: Read procedure byte 0x%.2x\n", dev->procbyte);
		    i=0;	/* resettimeout */
		  }
		}
		if ((dev->proto == 0) && (inb(iobase) & 0x08)){
		  DEBUG(4, "cardman_cs: T0Done flag set (read reply)\n");
		  break;
		}
		if (dev->proto) infolen=inb(iobase+1);
	}
	if (i==200) {
		DEBUG(4, "cardman_cs: timeout waiting for numRecBytes\n");
		rc=-EIO;
		goto release_io;
	}else{
	  if (dev->proto == 0){
	    /* DEBUG(4, "cardman_cs: Wait for T0Done bit to be  set\n");
	    for(i=0;i<1000;i++) {
	      if (inb(iobase)&0x08)
		break;
	      ipause(T_10MSEC);
	    }
	    if (i==1000) {
	      DEBUG(4, "cardman_cs: timeout waiting for T0Done\n");
	      rc=-EIO;
	      goto release_io;
	    }
	  */
	    dev->procbyte=inb(iobase+1);
	    DEBUG(4, "cardman_cs: Read procedure byte 0x%.2x\n", dev->procbyte);
	    IO_ReadNumRecBytes(iobase,&dev->rlen);
	    DEBUG(4, "cardman_cs: Read NumRecBytes = %i\n", dev->rlen);
	    IO_ReadNumRecBytes(iobase,&dev->rlen);
	    DEBUG(4, "cardman_cs: Reread NumRecBytes = %i\n", dev->rlen);
	  }
	}
	/* T=1: read offset=zero, T=0: read offset=after challenge */
	dev->rpos=dev->proto?0:nr==4?5:nr>dev->rlen?5:nr;
	DEBUG(4, "cardman_cs: dev->rlen = %i,  dev->rpos = %i, nr = %i\n", dev->rlen, dev->rpos, nr);

	/* last check before exit */
	/*
	  if (!IO_DetectCMM(iobase,dev))
		rc=-ENODEV;
	*/

release_io:
	DEBUG(4, "cardman_cs: Reset SM\n");
	xoutb(0x80,iobase);		/* reset SM */

	if (rc<0){
	  DEBUG(4, "Write failed but clear T_Active\n");
	  dev->flags1 &= 0xdf;
	  xoutb(dev->flags1,
		iobase+1);
	}

	//	if (rc<0)
	//	xoutb(0x80,iobase);		/* reset SM */
	clear_bit(LOCK_IO,&dev->flags);
	wake_up_interruptible(&dev->ioq);
	wake_up_interruptible(&dev->readq);	/* tell read we have data */

	/* ITSEC E2: clear write buffer */
	if (verify_area(VERIFY_WRITE,(void *)buf,count)==0)
		memset((char *)buf,0,count);

	/* return error or actually written bytes */
	DEBUG(2, "cardman_cs: <- cmm_write\n");
	return rc<0?rc:nr;
}

static void start_monitor(cardman_dev_t *dev) {
    DEBUG(3, "cardman_cs: -> start_monitor\n");
    if (!dev->monitor_running) {
      DEBUG(5, "cardman_cs: create, init and add timer\n");
      init_timer(&dev->timer);
      dev->monitor_running=1;
      dev->timer.expires=jiffies;
      dev->timer.data=(u_long)dev;
      dev->timer.function=monitor_card;
      add_timer(&dev->timer);
    }else   DEBUG(5, "cardman_cs: monitor already running\n");
    DEBUG(3, "cardman_cs: <- start_monitor\n");
}

static void stop_monitor(cardman_dev_t *dev) {
  DEBUG(3, "cardman_cs: -> stop_monitor\n");
  if (dev->monitor_running) {
    DEBUG(5, "cardman_cs: stopping monitor\n");
    terminate_monitor(dev);
    /* reset monitor SM */
    clear_bit(IS_ATR_VALID, &dev->flags);
    clear_bit(IS_ATR_PRESENT, &dev->flags);
  }
  else DEBUG(5, "cardman_cs: monitor already stopped\n");
  DEBUG(3, "cardman_cs: <- stop_monitor\n");
}

static int cmm_ioctl(struct inode *inode,struct file *filp,unsigned int cmd,
			unsigned long arg) {
register cardman_dev_t *dev=(cardman_dev_t *)filp->private_data;
register int iobase=dev->link.io.BasePort1;
dev_link_t *link;
int size;
int rc;
#ifdef PCMCIA_DEBUG
char *ioctl_names[CM_IOC_MAXNR+1]={
	[_IOC_NR(CM_IOCGSTATUS)] 	  "CM_IOCGSTATUS",
	[_IOC_NR(CM_IOCGATR)]		  "CM_IOCGATR",
	[_IOC_NR(CM_IOCARDOFF)]           "CM_IOCARDOFF",
        [_IOC_NR(CM_IOCSPTS)]             "CM_IOCSPTS",
	[_IOC_NR(CM_IOSDBGLVL)]           "CM4000_DBGLVL",
	[_IOC_NR(CM4000_IOCMONITOR)]	  "CM4000_IOCMONITOR",
	[_IOC_NR(CM4000_IOCDUMPATR)]	  "CM4000_IOCDUMPATR",
	[_IOC_NR(CM4000_IOCDECUSECOUNT)]  "CM4000_IOCDECUSECOUNT",
	[_IOC_NR(CM4000_IOCPOWERON)]	  "CM4000_IOCPOWERON",
	[_IOC_NR(CM4000_IOCGIOADDR)]	  "CM4000_IOCGIOADDR",
};
#endif
	DEBUG(2,"cardman_cs: cmm_ioctl(device=%d.%d process=%s,%d)\n",
		MAJOR(inode->i_rdev),MINOR(inode->i_rdev),
		dev->owner->comm,dev->owner->pid);

	link=dev_table[MINOR(inode->i_rdev)];
	if (!(DEV_OK(link))){
	  DEBUG(4, "cardman_cs: DEV_OK false\n");
		return -ENODEV;
	}
#ifdef PARANOIA
	if (dev==NULL){
	  DEBUG(4, "cardman_cs: dev== NULL\n");
		return -ENODEV;
	}
#endif
	if (test_bit(IS_CMM_ABSENT,&dev->flags)){
	  DEBUG(4, "cardman_cs: CMM_ABSENT flag set\n");
		return -ENODEV;
	}

	if (_IOC_TYPE(cmd)!=CM_IOC_MAGIC) {
	  DEBUG(4,"cardman_cs: ioctype mismatch\n");
	  return -EINVAL;
	}
	if (_IOC_NR(cmd)>CM_IOC_MAXNR) {
	  DEBUG(4,"cardman_cs: iocnr mismatch\n");
	  return -EINVAL;
	} 
	size=_IOC_SIZE(cmd);
	rc=0;
	DEBUG(4,"cardman_cs: iocdir=%.4x iocr=%.4x iocw=%.4x iocsize=%d cmd=%.4x\n",
		_IOC_DIR(cmd),_IOC_READ,_IOC_WRITE,size,cmd);

	DEBUG(4,"cardman_cs: %s\n",ioctl_names[_IOC_NR(cmd)]);
	/* I thought that with elf, null-pointed pages are valid? */

	if (_IOC_DIR(cmd)&_IOC_READ) {
		if (arg==0) rc=-EFAULT;
		else rc=verify_area(VERIFY_WRITE,(void *)arg,size);
	}
	if (_IOC_DIR(cmd)&_IOC_WRITE) {
		if (arg==0) rc=-EFAULT;
		else rc=verify_area(VERIFY_READ,(void *)arg,size);
	}
	if (rc)
		return rc;

	switch(cmd) {
	case CM_IOCGSTATUS:
	        DEBUG(4, "cardman_cs:  ... in CM_IOCGSTATUS\n");
		/* clear other bits, but leave inserted & powered as they are */
		*(unsigned int *)arg=dev->flags0 & 3;
		if (test_bit(IS_ATR_PRESENT,&dev->flags))
			*(unsigned int *)arg |= CM_ATR_PRESENT;
		if (test_bit(IS_ATR_VALID,&dev->flags))
			*(unsigned int *)arg |= CM_ATR_VALID;
		if (test_bit(IS_CMM_ABSENT,&dev->flags))
			*(unsigned int *)arg |= CM_NO_READER;
		if (test_bit(IS_BAD_CARD,&dev->flags))
			*(unsigned int *)arg |= CM_BAD_CARD;
		return 0;
	case CM_IOCGATR:
	        DEBUG(4, "cardman_cs:  ... in CM_IOCGATR\n");
		{
		register atreq_t *atreq=(atreq_t *)arg;
		int tmp;
		/* allow nonblocking io and being interrupted */
		while(test_bit(IS_ATR_PRESENT,&dev->flags)==0) {
			if (filp->f_flags & O_NONBLOCK)
				return -EAGAIN;
			interruptible_sleep_on(&dev->atrq);
			if (signal_pending(current))
				return -ERESTARTSYS;
		}
		if (test_bit(IS_ATR_VALID,&dev->flags)==0){
		  //atreq->atr_len=-1;
		  tmp = -1;
		  copy_to_user(&(atreq->atr_len), &tmp, sizeof(int));
		}else {
		  copy_to_user(atreq->atr, dev->atr, dev->atr_len);	
		  tmp = dev->atr_len;
		  copy_to_user(&(atreq->atr_len), &tmp, sizeof(int));
		}
		return 0;
		}
	case CM_IOCARDOFF:
	        
                #ifdef PCMCIA_DEBUG
	          DEBUG(4, "cardman_cs:  ... in CM_IOCARDOFF\n"); 
		  DEBUG(4, "cardman_cs: Card Status(dev->flags0):\n");
		  if (dev->flags0 & 0x01){ DEBUG(4, "    Card inserted\n");}else{ DEBUG(1, "    No card inserted\n");}
		  if (dev->flags0 & 0x02){ DEBUG(4, "    Card powered\n" );}else{ DEBUG(1, "    Card not powered\n");}
                #endif //PCMCIA_DEBUG

	        // is a card inserted and powered?
		if ((dev->flags0 & 0x01) && (dev->flags0 & 0x02)) {
		  
		  // get IO lock
		  while (test_and_set_bit(LOCK_IO, &dev->flags) == 1){
		    if (filp->f_flags & O_NONBLOCK)
		      return -EAGAIN;
		    interruptible_sleep_on(&dev->ioq);
		    if (signal_pending(current))
		      return -ERESTARTSYS;
		  }

		  // Set Flags0 = 0x42
		  DEBUG(4,"cardman_cs: Set Flags0=0x42 \n");
		  xoutb (0x42,iobase);
		  //dev->flags&=0x00000002; 
		  clear_bit(LOCK_IO, &dev->flags);
		  clear_bit(IS_ATR_VALID, &dev->flags); 
		  dev->mstate=M_CARDOFF;  
		  while(test_bit(IS_ATR_VALID,&dev->flags)==0) {
		    if (filp->f_flags & O_NONBLOCK)
		      return -EAGAIN;
		    interruptible_sleep_on(&dev->atrq);
		    if (signal_pending(current))
		      return -ERESTARTSYS;
		  }
		}

		//release lock
		clear_bit(LOCK_IO, &dev->flags);
		wake_up_interruptible(&dev->ioq);
		
		return 0;
        case CM_IOCSPTS:
                {
                  ptsreq_t *ptsreq=(ptsreq_t *)arg;
		  		  		  
		  rc = 0;
                  DEBUG(4, "cardman_cs: ... in CM_IOCSPTS\n");
                  // wait for ATR to get valid
                  while(test_bit(IS_ATR_PRESENT,&dev->flags)==0) {
			if (filp->f_flags & O_NONBLOCK)
				return -EAGAIN;
			interruptible_sleep_on(&dev->atrq);
			if (signal_pending(current))
				return -ERESTARTSYS;
		  }
		  // get IO lock
		  while (test_and_set_bit(LOCK_IO, &dev->flags) == 1){
		    if (filp->f_flags & O_NONBLOCK)
		      return -EAGAIN;
		    interruptible_sleep_on(&dev->ioq);
		    if (signal_pending(current))
		      return -ERESTARTSYS;
		  }
		  if((rc=set_protocol(dev, ptsreq))!=0){
		    // auto power_on again
		    dev->mstate = M_FETCH_ATR;
		    clear_bit(IS_ATR_VALID, &dev->flags);
		  }
		  //release lock
		  clear_bit(LOCK_IO, &dev->flags);
		  wake_up_interruptible(&dev->ioq);
                  
                }	        
                return rc;
#ifdef PCMCIA_DEBUG
	case CM_IOSDBGLVL: /* set debug log level*/
	        {
		  int old_pc_debug=0;

		  old_pc_debug = pc_debug;
		  pc_debug = *(int*)arg;
		  if (old_pc_debug != pc_debug)
		    DEBUG(0, "cardman_cs: Changed debug log level to %i\n", pc_debug);
		}
	        return rc;
	case CM4000_IOCMONITOR:	/* for debugging */
		if (arg) start_monitor(dev);
		else stop_monitor(dev);
		return 0;
	case CM4000_IOCDUMPATR:	/* for debugging */
		dev_dumpatr(link->priv);
		return 0;
	case CM4000_IOCDECUSECOUNT:	/* for debugging */
		MOD_DEC_USE_COUNT;
		return 0;
	case CM4000_IOCPOWERON:
		// cmm_poweron(link->io.BasePort1);
		return 0;
	/* needed by debugging tool only. still not official */
	case CM4000_IOCGIOADDR:
	        DEBUG(4, "cardman_cs:  ... in CM_IOCGIOADDR (needed for debuging\n");
		*(unsigned long *)arg=link->io.BasePort1;
		return 0;
#endif
	default:
                DEBUG(4, "cardman_cs:  ... in default (unknown IOCTL code)\n");
		return -EINVAL;
	}
}

static int cmm_open (struct inode *inode, struct file *filp) {
cardman_dev_t *dev;
dev_link_t *link;
unsigned long flags;
int i,rc;

	DEBUG(2,"cardman_cs: -> cmm_open(device=%d.%d process=%s,%d)\n",
		MAJOR(inode->i_rdev),MINOR(inode->i_rdev),
		current->comm,current->pid);

	i=MINOR(inode->i_rdev);
	if (i>=CM4000_MAX_DEV){
	  DEBUG(4, "cardman_cs: cardman_cs: MAX_DEV reached\n");
	  return -ENODEV;
	}
	link=dev_table[MINOR(inode->i_rdev)];
	if (link==NULL || !(DEV_OK(link))){
	  DEBUG(4, "cardman_cs: link== NULL || DEV_OK false\n");
	  return -ENODEV;
	}
	if (link->open){
	  DEBUG(4, "cardman_cs: DEVICE BUSY\n");
	  return -EBUSY;
	}

	dev=(cardman_dev_t *)link->priv;
	filp->private_data=dev;

	/* init device variables, they may be "polluted" after close
	 * or, the device may never have been closed (i.e. open failed)
	 */

	ZERO_DEV(dev);

	/* opening will always block since the
	 * monitor will be started by open, which
	 * means we have to wait for ATR becoming
	 * vaild = block until valid (or card
	 * inserted)
	 */
	if (filp->f_flags & O_NONBLOCK)
		return -EAGAIN;

	dev->owner=current;
	dev->mdelay=T_50MSEC;

	/* start monitoring the cardstatus */

	save_flags(flags);
	sti();

	/*
	init_timer(&dev->timer);
#ifdef PCMCIA_DEBUG
	dev->monitor_running=1;
#endif
	dev->timer.expires=jiffies;
	dev->timer.data=(u_long)dev;
	dev->timer.function=monitor_card;
	add_timer(&dev->timer);
	*/
	start_monitor(dev);

	link->open=1;		/* only one open per device */
	MOD_INC_USE_COUNT;
	rc=0;
#if 0
	/* wait until the monitor has fetched the atr
	 * for us. this can also mean that we have to
	 * wait for the user to insert a card
	 */
	interruptible_sleep_on(&dev->atrq);
	if (test_bit(IS_BAD_CSUM,&dev->flags)) {
		terminate_monitor(dev);
		rc=-EIO;
	}
	else if (signal_pending(current)) {
		terminate_monitor(dev);
		rc=-ERESTARTSYS;
	}
	restore_flags(flags);
	if (rc<0) {
		link->open=0;
		MOD_DEC_USE_COUNT;
	}
#endif
	DEBUG(2, "cardman_cs: <- cmm_open\n");
	return rc;
}

static int cmm_close(struct inode *inode,struct file *filp) {
cardman_dev_t *dev;
dev_link_t *link;
int i;

	DEBUG(2,"cardman_cs: -> cmm_close(maj/min=%d.%d)\n",
		MAJOR(inode->i_rdev),MINOR(inode->i_rdev));

	i=MINOR(inode->i_rdev);
	if (i>=CM4000_MAX_DEV)
		return -ENODEV;
	link=dev_table[MINOR(inode->i_rdev)];
	if (link==NULL)
		return -ENODEV;

	dev=(cardman_dev_t *)link->priv;
	stop_monitor(dev);

	ZERO_DEV(dev);

	link->open=0;		/* only one open per device */
	wake_up(&dev->devq);	/* socket removed? */
	MOD_DEC_USE_COUNT;
	DEBUG(2, "cardman_cs: cmm_close\n");
	return 0;
}

/*
 * called from pcmcia glue layer when
 * the reader is removed
 */

static void cmm_cardman_release(dev_link_t *link) {
cardman_dev_t *dev=(cardman_dev_t *)link->priv;

	/* dont terminate the monitor, rather rely on
	 * close doing that for us.
	 */
        DEBUG(3, "cardman_cs: -> cmm_cardman_release\n"); 
	while(link->open) {
		printk(KERN_INFO MODULE_NAME ": delaying release until "
			"process '%s', pid %d has terminated\n",
			dev->owner->comm,dev->owner->pid);
		/* note: don't interrupt us:
		 * close the applications which own
		 * the devices _first_ !
		 */
		sleep_on(&dev->devq);
	}
	// dev->devq=NULL;		/* this cannot be zeroed earlier */
	DEBUG(3, "cardman_cs: <- cmm_cardman_release\n"); 
	return;
}

/*==== Interface to PCMCIA Layer =======================================*/

static void cs_error(client_handle_t handle, int func, int ret) {
error_info_t err = { func, ret };

	CardServices(ReportError, handle, &err);
}

/*================+
 | cardman_config |
 +================*/

#define	CS_CHECK(fn,args...) \
	((fail_rc=CardServices(fail_fn=(fn),args))!=CS_SUCCESS)

static void cardman_config(dev_link_t *link,int devno) {
client_handle_t handle;
cardman_dev_t *dev;
tuple_t tuple;
cisparse_t parse;
config_info_t conf;
u_char buf[64];
int fail_fn,fail_rc;
int rc;

	DEBUG(3,"cardman_cs: -> cardman_config\n");

	/* read the config-tuples */

	handle=link->handle;

	tuple.DesiredTuple=CISTPL_CONFIG;
	tuple.Attributes=0;
	tuple.TupleData=buf;
	tuple.TupleDataMax=sizeof(buf);
 	tuple.TupleOffset=0;

	if (CS_CHECK(GetFirstTuple,handle,&tuple))
		goto cs_failed;
	if (CS_CHECK(GetTupleData,handle,&tuple))
		goto cs_failed;
	if (CS_CHECK(ParseTuple,handle,&tuple,&parse))
		goto cs_failed;
	if (CS_CHECK(GetConfigurationInfo,handle,&conf))
		goto cs_failed;

	link->state|=DEV_CONFIG;
	link->conf.ConfigBase=parse.config.base;
	link->conf.Present=parse.config.rmask[0];
	link->conf.Vcc=conf.Vcc;
	DEBUG(5,"cardman_cs: link->conf.Vcc=%d\n",link->conf.Vcc);

	link->io.BasePort2=0;
	link->io.NumPorts2=0;
	link->io.Attributes2=0;
	tuple.DesiredTuple=CISTPL_CFTABLE_ENTRY;
	for(rc=CardServices(GetFirstTuple,handle,&tuple);
		rc==CS_SUCCESS;
		rc=CardServices(GetNextTuple,handle,&tuple)) {

 	        DEBUG(5, "cardman_cs: Examing CIS Tuple!\n");

		rc=CardServices(GetTupleData,handle,&tuple);
		if(rc!=CS_SUCCESS)
			continue;
		rc=CardServices(ParseTuple,handle,&tuple,&parse);
		if(rc!=CS_SUCCESS)
			continue;

		DEBUG(5,"cardman_cs: tupleIndex=%d\n",parse.cftable_entry.index);
		link->conf.ConfigIndex=parse.cftable_entry.index;

		/* Get the IOaddr */
		if (parse.cftable_entry.io.nwin) {
			link->io.BasePort1=parse.cftable_entry.io.win[0].base;
			link->io.NumPorts1=parse.cftable_entry.io.win[0].len;
			link->io.Attributes1=IO_DATA_PATH_WIDTH_AUTO;
			if(!(parse.cftable_entry.io.flags & CISTPL_IO_8BIT))
				link->io.Attributes1=IO_DATA_PATH_WIDTH_16;
			if(!(parse.cftable_entry.io.flags & CISTPL_IO_16BIT))
				link->io.Attributes1=IO_DATA_PATH_WIDTH_8;
#ifdef PCMCIA_DEBUG
			if (io) link->io.BasePort1=io;
#endif
			link->io.IOAddrLines=parse.cftable_entry.io.flags
						& CISTPL_IO_LINES_MASK;
			DEBUG(5,"cardman_cs: io.BasePort1=%.4x\n",link->io.BasePort1);
			DEBUG(5,"cardman_cs: io.NumPorts1=%.4x\n",link->io.NumPorts1);
			DEBUG(5,"cardman_cs: io.BasePort2=%.4x\n",link->io.BasePort2);
			DEBUG(5,"cardman_cs: io.NumPorts2=%.4x\n",link->io.NumPorts2);
			DEBUG(5,"cardman_cs: io.IOAddrLines=%.4x\n",link->io.IOAddrLines);
			rc=CardServices(RequestIO,handle,&link->io);
			if (rc==CS_SUCCESS) {
				DEBUG(5,"cardman_cs: RequestIO OK\n");
				break; /* we are done */
			}
			else DEBUG(5,"cardman_cs: RequestIO failed\n");
		}
	}
	if (rc!=CS_SUCCESS) {
		DEBUG(5,"cardman_cs: Couldn't configure cardman\n");
		goto cs_release;
	}

        link->conf.IntType=00000002;

	if (CS_CHECK(RequestConfiguration,handle,&link->conf))
		goto cs_release;
	DEBUG(5, "cardman_cs: RequestConfiguration OK\n");

	dev=link->priv;
	sprintf(dev->node.dev_name,DEVICE_NAME "%d",devno);
	dev->node.major=major;
	dev->node.minor=devno;
	dev->node.next=NULL;
	link->dev=&dev->node;
	link->state&=~DEV_CONFIG_PENDING;

	DEBUG(5,"cardman_cs: device " DEVICE_NAME "%d at 0x%.4x-0x%.4x\n",
		devno,
		link->io.BasePort1,
		link->io.BasePort1+link->io.NumPorts1);

	DEBUG(5, "cardman_cs: <- cardman_config (succ)\n");
	return;

cs_failed:
	cs_error(handle,fail_fn,fail_rc);
cs_release:
	cardman_release((u_long)link);
	link->state&=~DEV_CONFIG_PENDING;
	DEBUG(5, "cardman_cs: <- cardman_config (failure)\n");
}

/*===============+
 | cardman_event |
 +===============*/

static int cardman_event(event_t event, int priority,
			event_callback_args_t *args) {
dev_link_t *link;
cardman_dev_t *dev;
int devno;

	DEBUG(3,"cardman_cs: -> cardman_event\n");
	link=args->client_data;
	dev=link->priv;
	for(devno=0;devno<CM4000_MAX_DEV;devno++)
		if (dev_table[devno]==link)
			break;
	if (devno==CM4000_MAX_DEV)
		return CS_BAD_ADAPTER;
	switch(event) {
	case CS_EVENT_CARD_INSERTION:
		DEBUG(5, "cardman_cs: CS_EVENT_CARD_INSERTION\n");
		link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
		cardman_config(link,devno);
		break;
	case CS_EVENT_CARD_REMOVAL:
		DEBUG(5, "cardman_cs: CS_EVENT_CARD_REMOVAL\n");
		link->state &= ~DEV_PRESENT;
		stop_monitor(dev);
		break;
	case CS_EVENT_PM_SUSPEND:
		DEBUG(5, "cardman_cs: CS_EVENT_PM_SUSPEND (fall-through to CS_EVENT_RESET_PHYSICAL)\n");
		link->state |= DEV_SUSPEND;
		/* fall-through */
	case CS_EVENT_RESET_PHYSICAL:
		DEBUG(5, "cardman_cs: CS_EVENT_RESET_PHYSICAL\n");
		if (link->state & DEV_CONFIG){
		  DEBUG(5, "cardman_cs: ReleaseConfiguration\n");
		  CardServices(ReleaseConfiguration, link->handle);
		}
		stop_monitor(dev);
		break;
	case CS_EVENT_PM_RESUME:
		DEBUG(5, "cardman_cs: CS_EVENT_PM_RESUME (fall-through to CS_EVENT_CARD_RESET)\n");
		link->state &= ~DEV_SUSPEND;
		/* fall-through */
	case CS_EVENT_CARD_RESET:
	        DEBUG(5, "cardman_cs: CS_EVENT_CARD_RESET\n");
		if ((link->state & DEV_CONFIG)){
		  DEBUG(5, "cardman_cs: RequestConfiguration\n");
		  CardServices(RequestConfiguration, link->handle, &link->conf);
		}
		if (link->open)
	          start_monitor(dev);
		break;
	default:
		DEBUG(5, "cardman_cs: cardman_event: unknown event %.2x\n",
			event);
		break;
	}
	DEBUG(3, "cardman_cs: <- cardman_event\n");
	return CS_SUCCESS;
}
/*=================+
 | cardman_release |
 +=================*/

static void cardman_release(u_long arg) {
dev_link_t *link;
int rc;

	DEBUG(3,"cardman_cs: -> cardman_release\n");
	link=(dev_link_t *)arg;
	cmm_cardman_release(link->priv); // delay release until device is closed
	rc=CardServices(ReleaseConfiguration,link->handle);
	if (rc!=CS_SUCCESS)
		DEBUG(5,"cardman_cs: couldn't ReleaseConfiguration reasoncode=%.2x\n",rc);
	rc=CardServices(ReleaseIO,link->handle,&link->io);
	if (rc!=CS_SUCCESS)
		DEBUG(5,"cardman_cs: couldn't ReleaseIO reasoncode=%.2x\n",rc);
	DEBUG(3,"cardman_cs: <- cardman_release\n");
}

/*================+
 | cardman_attach |
 +================*/

static dev_link_t *cardman_attach(void) {
cardman_dev_t *dev;
dev_link_t *link;
client_reg_t client_reg;
int i;

	DEBUG(3,"cardman_cs: cardman_attach\n");
	for(i=0;i<CM4000_MAX_DEV;i++)
		if (dev_table[i]==NULL)
			break;
	if (i==CM4000_MAX_DEV) {
		printk(KERN_NOTICE "cardman_cs: all devices in use\n");
		return NULL;
	}
	/* create a new cardman_cs device */
	DEBUG(5, "cardman_cs: create cardman device instance\n");
	dev=kmalloc(sizeof(cardman_dev_t), GFP_KERNEL);
	if (dev==NULL) return NULL;
	memset(dev,0,sizeof(cardman_dev_t));
	link=&dev->link; link->priv=dev;

	link->release.function=&cardman_release;
	link->release.data=(u_long)link;
	link->conf.IntType=INT_MEMORY_AND_IO;
	dev_table[i]=link;

	/* register with card services */
	DEBUG(5, "cardman_cs: Register with Card Services\n");
	client_reg.dev_info=&dev_info;
	client_reg.Attributes=INFO_IO_CLIENT | INFO_CARD_SHARE;
	client_reg.EventMask=
		CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
		CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
		CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
	client_reg.event_handler=&cardman_event;
	client_reg.Version=0x0210;
	client_reg.event_callback_args.client_data=link;
	i=CardServices(RegisterClient,&link->handle,&client_reg);
	if (i) {
		cs_error(link->handle,RegisterClient,i);
		cardman_detach(link);
		return NULL;
	}
	init_waitqueue_head(&dev->devq);
	init_waitqueue_head(&dev->ioq);
	init_waitqueue_head(&dev->atrq);
	init_waitqueue_head(&dev->readq);
	return link;
}

/*================+
 | cardman_detach |
 +================*/

static void cardman_detach_by_devno(int devno,dev_link_t *link) {
cardman_dev_t *dev=link->priv;

	DEBUG(3,"cardman_cs: -> detach_by_devno(devno=%d)\n",devno);
	if (link->state & DEV_CONFIG){
	  DEBUG(5, "cardman_cs: device still configured (try to release it)\n");
	  cardman_release((u_long)link);
	}
	if (link->handle) CardServices(DeregisterClient,link->handle);
	dev_table[devno]=NULL;
	DEBUG(5,"cardman_cs: freeing dev=%.8x\n",(unsigned int)dev);
	kfree(dev);
	DEBUG(3, "cardman_cs: <- detach_by-devno\n");
	return;
}

static void cardman_detach(dev_link_t *link) {
int i;
	DEBUG(3,"cardman_cs: -> cardman_detach(link=%.8x\n",
	 	(unsigned int)link);
	/* find device */
	for(i=0;i<CM4000_MAX_DEV;i++)
		if (dev_table[i]==link)
			break;
	if (i==CM4000_MAX_DEV) {
		printk(KERN_WARNING "cardman_cs: detach for unkown device aborted\n");
		return;
	}
	cardman_detach_by_devno(i,link);
	DEBUG(3, "cardman_cs: <- cardman_detach\n");
	return;
}

/*==============+
 | init_cardman |
 +==============*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static int __init cmm_init(void) {
#else
int init_module(void) {
#endif
#ifdef CHECK_VERSION
servinfo_t serv;
#endif

	printk(KERN_INFO "%s\n",version);
#ifdef CHECK_VERSION
	CardServices(GetCardServicesInfo, &serv);
	if (serv.Revision != CS_RELEASE_CODE) {
		printk(KERN_NOTICE "cardman_cs: Card Services release "
			"does not match.\n");
		return -1;
	}
#endif
	register_pccard_driver(&dev_info, &cardman_attach, &cardman_detach);
	major=register_chrdev(0,DEVICE_NAME,&cardman_fops);
	if (major<0) {
		printk(KERN_WARNING "cardman_cs: could not get major number\n");
		return -1;
	}
	return 0;
}

/*==============+
 | exit_cardman |
 +==============*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void __exit cmm_exit(void) {
#else
void cleanup_module(void) {
#endif
int i;
	printk(KERN_INFO "cardman_cs unloading\n");
	unregister_pccard_driver(&dev_info);
	for(i=0;i<CM4000_MAX_DEV;i++)
		if (dev_table[i]) cardman_detach_by_devno(i,dev_table[i]);
	 unregister_chrdev(major,DEVICE_NAME);
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
module_init(cmm_init);
module_exit(cmm_exit);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("Dual BSD/GPL");
#endif
